import numpy as np


def pitch_weight(pitch):
    return -pitch / 128 + 1


def velocity_weight(velocity):
    return velocity / 128


def notes_weight(notes):
    pitches = np.array([note.pitch for note in notes])
    velocities = np.array([note.velocity for note in notes])

    return pitch_weight(pitches) + velocity_weight(velocities)


def get_heat_map(notes, n_bins=24, beat_resolution=480, is_drum=False, pitches=range(0, 128)):
    bin_resolution = np.ceil(beat_resolution / n_bins).astype(int)
    notes = [note for note in notes if note.pitch in pitches and note.instrument.is_drum == is_drum]
    heat_map = np.zeros(n_bins)
    weights = notes_weight(notes)
    if is_drum:
        weights = np.clip(np.ceil(weights), 0., 1.)
    for note, weight in zip(notes, weights):
        heat_map[(note.start % beat_resolution) // bin_resolution] += weight

    return heat_map



from .heat_map import get_heat_map
from .melody_extraction import extract_accompaniment
from .timepoints import get_notes_by_period
import numpy as np

drum_names_to_pitches = {
    'bass_drum': [35, 36],
    'closed_hihat': [42],
    'crash': [49, 57],
    'floor_tom': [41, 43],
    'open_hihat': [46],
    'pedal_hihat': [44],
    'ride': [51, 59],
    'snare': [38, 39, 40],
    'tambourine': [54],
    'tom': [45, 47, 48, 50]
}


def add_instrument_back_pointer(midi_obj):
    for instrument in midi_obj.instruments:
        for note in instrument.notes:
            note.instrument = instrument


def get_heat_maps(midi_obj, n_bins=24, beat_resolution=480, rid_melody=False, is_drum=False, pitches=range(0, 128)):
    add_instrument_back_pointer(midi_obj)
    notes = []
    for instrument in midi_obj.instruments:
        notes += instrument.notes
    if rid_melody:
        notes = extract_accompaniment(notes)
    notes_by_beats = get_notes_by_period(notes, resolution=beat_resolution)

    return np.stack([
        get_heat_map(
            notes=note_set,
            n_bins=n_bins,
            beat_resolution=beat_resolution,
            is_drum=is_drum,
            pitches=pitches
        )
        for note_set in notes_by_beats])


def get_drum_maps(midi_obj, n_bins=96, bar_resolution=1920, drums=None):
    if drums is None:
        drums = ['bass_drum', 'closed_hihat', 'snare']

    return np.stack([get_heat_maps(
        midi_obj=midi_obj,
        n_bins=n_bins,
        beat_resolution=bar_resolution,
        is_drum=True,
        pitches=drum_names_to_pitches[drum]
    ) for drum in drums], axis=1)


def get_dataset(midi_objs, n_bins=24, beat_resolution=480, rid_melody=False, is_drum=False, pitches=range(0, 128)):
    dataset = np.zeros((0, n_bins))
    for midi_obj in midi_objs:
        heat_maps = get_heat_maps(
            midi_obj=midi_obj,
            n_bins=n_bins,
            beat_resolution=beat_resolution,
            rid_melody=rid_melody,
            is_drum=is_drum,
            pitches=pitches)
        dataset = np.concatenate((dataset, heat_maps))

    return dataset


def get_drum_dataset(midi_objs, n_bins=96, drums=None):
    if drums is None:
        drums = ['bass_drum', 'closed_hihat', 'snare']

    dataset = np.zeros((0, len(drums), n_bins))
    for midi_obj in midi_objs:
        time_signatures = midi_obj.time_signature_changes
        if len(time_signatures) > 1:
            continue
        num = time_signatures[0].numerator
        while num % 2 == 0:
            num /= 2
        if num != 1:
            continue

        drum_maps = get_drum_maps(
            midi_obj=midi_obj,
            n_bins=n_bins,
            bar_resolution=midi_obj.ticks_per_beat * 4,
            drums=drums
        )
        dataset = np.concatenate((dataset, drum_maps))

    return dataset



import numpy as np


def quantize_melody(notes, tick_resol=240):
    melody_notes = []
    for note in notes:
        # cut too long notes
        if note.end - note.start > tick_resol * 8:
            note.end = note.start + tick_resol * 4

        # quantize
        note.start = int(np.round(note.start / tick_resol) * tick_resol)
        note.end = int(np.round(note.end / tick_resol) * tick_resol)

        # append
        melody_notes.append(note)
    return melody_notes


def extract_melody(notes):
    # quanrize
    melody_notes = quantize_melody(notes)

    # sort by start, pitch from high to low
    melody_notes.sort(key=lambda x: (x.start, -x.pitch))

    # exclude notes < 60
    bins = []
    prev = None
    tmp_list = []
    for nidx in range(len(melody_notes)):
        note = melody_notes[nidx]
        if note.pitch >= 60:
            if note.start != prev:
                if tmp_list:
                    bins.append(tmp_list)
                tmp_list = [note]
            else:
                tmp_list.append(note)
            prev = note.start

            # preserve only highest one at each step
    notes_out = []
    for b in bins:
        notes_out.append(b[0])

    # avoid overlapping
    notes_out.sort(key=lambda x: x.start)
    for idx in range(len(notes_out) - 1):
        if notes_out[idx].end >= notes_out[idx + 1].start:
            notes_out[idx].end = notes_out[idx + 1].start

    # delete note having no duration
    notes_clean = []
    for note in notes_out:
        if note.start != note.end:
            notes_clean.append(note)

            # filtered by interval
    notes_final = [notes_clean[0]] if notes_clean != [] else []
    for i in range(1, len(notes_clean) - 1):
        if ((notes_clean[i].pitch - notes_clean[i - 1].pitch) <= -9) and \
                ((notes_clean[i].pitch - notes_clean[i + 1].pitch) <= -9):
            continue
        else:
            notes_final.append(notes_clean[i])
    notes_final += [notes_clean[-1]] if notes_clean != [] else []
    return notes_final


def extract_accompaniment(notes):
    melody = extract_melody(notes)
    note_set = set(notes)
    melody_set = set(melody)

    return list(note_set.difference(melody_set))



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="groover",
    version="0.1.0",
    author="Joshua Chang",
    author_email="chchang6@illinois.edu",
    description="A rhythm feature extractor and classifier for MIDI files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url="https://github.com/joshuachang2311/chorder",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        "numpy",
        "miditoolkit"
    ]
)
